#Exercise 3 template for CHE210D

#import python modules
import numpy as np
import time

#import compiled fortran library
import prolib

import sys
import os

import random as rd

import atomwrite

#whether or not to use the 3d visualization
UseVisual = False
if UseVisual:
    #import custom visualization library
    import atomvis


#========== GLOBAL VARIABLES ==========
#distance cutoff for pairwise interactions
Cut =10

#set the random number seed; useful for debugging


#NOTE:
#everything below assumes unit atomic masses,
#such that forces = accelerations.



def LineSearch(Pos, Dir, dx, EFracTol,
               Accel = 1.5, MaxInc = 10., MaxIter = 10000):
    """Performs a line search along direction Dir.
Input:
    Pos: starting positions, (N,3) array
    Dir: (N,3) array of gradient direction
    dx: initial step amount
    EFracTol: fractional energy tolerance
    Accel: acceleration factor
    MaxInc: the maximum increase in energy for bracketing
    MaxIter: maximum number of iteration steps
Output:
    PEnergy: value of potential energy at minimum along Dir
    Pos: minimum energy (N,3) position array along Dir
"""
    #identify global variables
    global M, L, Cut, eps, lam, Name
    
    #start the iteration counter
    Iter = 0

    #find the normalized direction    
    NormDir = np.clip(Dir, -1.e100, 1.e100)
    NormDir = NormDir / np.sqrt(np.sum(NormDir * NormDir))

    #take the first two steps and compute energies    
    Dists = [0., dx]
    PEs = [prolib.calcenergy(Pos + NormDir * x, M, L, Cut, eps, lam, Name) for x in Dists]
    
    #if the second point is not downhill in energy, back
    #off and take a shorter step until we find one
    while PEs[1] > PEs[0]:
        Iter += 1
        dx = dx * 0.5
        Dists[1] = dx
        PEs[1] = prolib.calcenergy(Pos + NormDir * dx, M, L, Cut, eps, lam, Name)
        
    #find a third point
    Dists = Dists + [2. * dx]
    PEs = PEs + [prolib.calcenergy(Pos + NormDir * 2. * dx, M, L, Cut, eps, lam, Name)]
    
    #keep stepping forward until the third point is higher
    #in energy; then we have bracketed a minimum
    while PEs[2] < PEs[1]:
        Iter += 1
            
        #find a fourth point and evaluate energy
        Dists = Dists + [Dists[-1] + dx]
        PEs = PEs + [prolib.calcenergy(Pos + NormDir * Dists[-1], M, L, Cut, eps, lam, Name)]

        #check if we increased too much in energy; if so, back off
        if (PEs[3] - PEs[0]) > MaxInc * (PEs[0] - PEs[2]):
            PEs = PEs[:3]
            Dists = Dists[:3]
            dx = dx * 0.5
        else:
            #shift all of the points over
            PEs = PEs[-3:]
            Dists = Dists[-3:]
            dx = dx * Accel
            
    #we've bracketed a minimum; now we want to find it to high
    #accuracy
    OldPE3 = 1.e300
    while True:
        Iter += 1
        if Iter > MaxIter:
            print "Warning: maximum number of iterations reached in line search."
            break
            
        #store distances for ease of code-reading
        d0, d1, d2 = Dists
        PE0, PE1, PE2 = PEs

        #use a parobolic approximation to estimate the location
        #of the minimum
        d10 = d0 - d1
        d12 = d2 - d1
        Num = d12*d12*(PE0-PE1) - d10*d10*(PE2-PE1)
        Dem = d12*(PE0-PE1) - d10*(PE2-PE1)
        if Dem == 0:
            #parabolic extrapolation won't work; set new dist = 0 
            d3 = 0
        else:
            #location of parabolic minimum
            d3 = d1 + 0.5 * Num / Dem
            
        #compute the new potential energy
        PE3 = prolib.calcenergy(Pos + NormDir * d3, M, L, Cut, eps, lam, Name)
        
        #sometimes the parabolic approximation can fail;
        #check if d3 is out of range < d0 or > d2 or the new energy is higher
        if d3 < d0 or d3 > d2 or PE3 > PE0 or PE3 > PE1 or PE3 > PE2:
            #instead, just compute the new distance by bisecting two
            #of the existing points along the line search
            if abs(d2 - d1) > abs(d0 - d1):
                d3 = 0.5 * (d2 + d1)
            else:
                d3 = 0.5 * (d0 + d1)
            PE3 = prolib.calcenergy(Pos + NormDir * d3, M, L, Cut, eps, lam, Name)
            
        #decide which three points to keep; we want to keep
        #the three that are closest to the minimum
        if d3 < d1:
            if PE3 < PE1:
                #get rid of point 2
                Dists, PEs = [d0, d3, d1], [PE0, PE3, PE1]
            else:
                #get rid of point 0
                Dists, PEs = [d3, d1, d2], [PE3, PE1, PE2]
        else:
            if PE3 < PE1:
                #get rid of point 0
                Dists, PEs = [d1, d3, d2], [PE1, PE3, PE2]
            else:
                #get rid of point 2
                Dists, PEs = [d0, d1, d3], [PE0, PE1, PE3]
                
        #check how much we've changed
        if abs(OldPE3 - PE3) < EFracTol * abs(PE3):
            #the fractional change is less than the tolerance,
            #so we are done and can exit the loop
            break
        OldPE3 = PE3

    #return the position array at the minimum (point 1)        
    PosMin = Pos + NormDir * Dists[1]
    PEMin = PEs[1]

    #if using visualization, update the display
    if UseVisual:
        if atomvis.Initialized:
            #update the positions
            atomvis.Update(PosMin)
        else:
            #initialize the visualization window
            atomvis.Init(PosMin)
        
    return PEMin, PosMin

        
def ConjugateGradient(Pos, dx, EFracTolLS, EFracTolCG):
    """Performs a conjugate gradient search.
Input:
    Pos: starting positions, (N,3) array
    dx: initial step amount
    EFracTolLS: fractional energy tolerance for line search
    EFracTolCG: fractional energy tolerance for conjugate gradient
Output:
    PEnergy: value of potential energy at minimum
    Pos: minimum energy (N,3) position array 
"""
    #identify global variables
    global M, L, Cut, eps, lam, Name
    #initial search direction
    Forces = np.zeros_like(Pos)
    PE, Forces, CollForces = prolib.calcenergyforces(Pos, M, L, Cut, eps, lam, Name, Forces)
    Dir = Forces
    OldPE = 1.e300
    #iterative line searches
    while abs(PE - OldPE) > EFracTolCG * abs(PE):
        OldPE = PE
        PE, Pos = LineSearch(Pos, Dir, dx, EFracTolLS)
        OldForces = Forces.copy()
        PE, Forces, CollForces = prolib.calcenergyforces(Pos, M, L, Cut, eps, lam, Name, Forces)
        Gamma = np.sum((Forces - OldForces) * Forces) / np.sum(OldForces * OldForces)
        Dir = Forces + Gamma *  Dir
    return PE, Pos


def InitPositions(N, L):
    """Returns an array of initial positions of each atom,
placed on a cubic lattice for convenience.
Input:
    N: number of atoms
    L: box length
Output:
    Pos: (N,3) array of positions
"""
    #make the position array
    Pos = np.zeros((N,3), float)
    #compute integer grid # of locations for cubic lattice
    NLat = int(N**(1./3.) + 1.)
    LatSpac = L / NLat
    #make an array of lattice sites
    r = LatSpac * np.arange(NLat, dtype=float) - 0.5*L
    #loop through x, y, z positions in lattice until done
    #for every atom in the system
    i = 0
    for x in r:
        for y in r:
            for z in r:
                Pos[i] = np.array([x,y,z], float)
                #add a random offset to help initial minimization
                Offset = 0.1 * LatSpac * (np.random.rand(3) - 0.5)
                Pos[i] = Pos[i] + Offset
                i += 1
                #if done placing atoms, return
                if i >= N:
                    return Pos
    return Pos

def RescaleVelocities(Vel, T):
    """Rescales velocities in the system to the target temperature.
Input:
    Vel: (N,3) array of atomic velocities
    T: target temperature
Output:
    Vel: same as above
"""
    #recenter to zero net momentum (assuming all masses same)
    Vel[2:] = Vel[2:] - Vel[2:].mean(axis=0)
    #find the total kinetic energy
    KE = 0.5 * np.sum(Vel * Vel)
    #find velocity scale factor from ratios of kinetic energy
    VScale = np.sqrt(1.5 * (len(Vel)-2) * T / KE)
    Vel[2:] = Vel[2:] * VScale
    Vel[0]=[0,0,0]
    Vel[1]=[0,0,0]
    return Vel  


def InitVelocities(N, T):
    """Returns an initial random velocity set.
Input:
    N: number of atoms
    T: target temperature
Output:
    Vel: (N,3) array of atomic velocities
"""
    Vel = np.random.rand(N, 3)
    Vel = RescaleVelocities(Vel, T)
    Vel = RescaleVelocities(Vel,T)
    Vel[0]=[0,0,0]
    Vel[1]=[0,0,0]
    return Vel


def InitAccel(Pos):
    """Returns the initial acceleration array.
Input:
    Pos: (N,3) array of atomic positions
Output:
    Accel: (N,3) array of acceleration vectors
"""
    #global variables
    global M, L, Cut, eps, lam, Name
    #get the acceleration from the forces
    Forces = np.zeros_like(Pos)
    PEnergy, Accel, CollForces = prolib.calcenergyforces(Pos, M, L, Cut, eps, lam, Name, Forces)
    Accel[0]=[0,0,0]
    Accel[1]=[0,0,0]
    return Accel


def InstTemp(Vel):
    """Returns the instantaneous temperature.
Input:
    Vel: (N,3) array of atomic velocities
Output:
    Tinst: float
"""
    return np.sum(Vel * Vel) / (3. * (len(Vel)-2))

def Andersen(Vel,T):
    """Uses the Andersen thermostat to rescale a particle's velocity according
        to the Maxwell-Boltzmann distribution.
Input:
    Vel: (N,3) array of atomic velocities
    T: Desired temperature
Output:
    Vel: (N,3) array of thermostatted atomic velocities"""
    
    global NMon
    
    stdev=np.sqrt(T)
    
    #Part=rd.randint(2,NMon+1)
    #Vel[Part,:]=np.random.normal(0,stdev,3)
    
    for Part in range(NMon+1)[2:]:
        Vel[Part]=np.random.normal(0,stdev,3)
    #print Vel
    
    return Vel


def EndtoEnd(Pos):
    """Returns squared end to end distance of polymer"""
    
    global NMon
    
    return np.sum((Pos[NMon+1]-Pos[2])*(Pos[NMon+1]-Pos[2]))


"""Module level code
        Runs a Molecular Dynamics simulation for a list of different polymer lengths MList
        using the Lennard-Jones chain model. Prints to a file the time, instantaneous temperature,
        potential energy, kinetic energy, and total energy. 
        3 steps
            1. Equilibration stage (NStepsEquil1 steps) with Andersen thermostat to desired temperature.
            2. Equilibration stage(NstepsEquil2 steps) with Andersen thermostat to desired temperature, at the end rescale
                velocities so instantaneous total energy equals average total energy.
            3. Production stage (NStepsProd steps) with no thermostat, track mean end to end distance and 
                force on colloids.
"""
        
"""Input MD Simulation parameters"""    
MList=[15]            #Polymer length to simulate
NPoly=1         #Number of polymers in simulation
dt=0.0001       #MD time step

#List of colloid separation distances to test
CollDistList=[10,12,14,16,18,20]

L=150           #Box side length

NTrials=3

NStepsEquil1    =   500000    #50s first equilibration
NStepsEquil2    =   500000    #50s second equilibration 
NStepsProd      =   1000000   #100s production run
PlotFreq=1000                   #how often to track temp, PE, etc for plots

"""Thermostat"""
T=1.0                 #Desired temperature
AndersenFreq=500       #Frequencey of Andersen thermostat collisions

"""Electrostatics"""
eps=10.0                 #Electrostatic strength 
lam=0.111               #1/Debye screening length

"""Movie"""
#Pdb=atomwrite.pdbfile("anim.pdb")   #Open PDB file to write positions
#MovieStartStep=10000                #Step during production phase when movie starts
#MovieStopStep=110000                 #Step during production phase when movie ends
#PdbFreq=200                         #Frequency of tracking positions for movie

"""Files"""
f=file("30 Force Run.txt","w")    #Open file to write data

"""Plotting"""
#import matplotlib.pyplot as plt
#colorlist=["g+","y+","r+","m+","c+","b+","y+","gs","m+"]    #list of colors for potential energy plot      
#colorlistindex=0                                            #indicates which color to use
#plt.figure()                                                #Open PE versus time figure

"""Initialize Lists"""
DistList=[]         #List of end to end polyelectrolyte distance
MeanForceList=[]    #List of Mean Force on colloid in x direction
MainTempList=[]     #List of all temperatures (to check thermostat)

#generate timelist for plotting things versus time
timelist=[]
for i in range(NStepsEquil1+NStepsEquil2+NStepsProd):
    if i%PlotFreq==0:
        timelist.append(dt*i)

print "Start MD"

"""Do MD simulation for each colloid separation distance"""
for M in MList:
    for CollDist in CollDistList:
        for NTrial in range(NTrials):
            NMon=M*NPoly            #Find number of monomers
            Name=[0,0]              #Make list of names - 0 for colloid and 1 for monomer
            for N in range(NMon):
                Name.append(1)
            NAtom=NMon+2
            
            PosColl=np.array([[-CollDist/2.,0,0],[CollDist/2.,0,0]])    #Place colloids at desired distance
            
            #Initialize Lists
            PEList=[]       #Potential energy
            TempList=[]     #Instantaneous Temperature
            EList=[]        #Total Energy
            
            Pos=InitPositions(NMon,L/100.)      #Initialize monomer positions between colloids
            Pos=np.concatenate((PosColl,Pos))   #Put colloid and monomer positions into one array
            PE,Pos=ConjugateGradient(Pos,0.001,1.e-8,1.e-10)    #Use conjugate gradient to minimize PE to start
            
            Vel=InitVelocities(NAtom,T)     #Get random initial velocities rescaled to desired temp.
            print InstTemp(Vel)             #Print temp. to check it is right
            Accel=InitAccel(Pos)            #Get initial accelerations based on forces
        
            """First Equilibration Run"""
            for step in range(NStepsEquil1):
                
                #Check if need to do Andersen thermostat
                if step%AndersenFreq==0:
                    Vel=Andersen(Vel,T)
                    #if step%10000==0:
                        #print InstTemp(Vel)     #print temp. to see if thermostat is working
                
                #Do an MD step        
                Pos, Vel, Accel, KEnergy, PEnergy, CollForces = prolib.vvintegrate(Pos, Vel, Accel, M, L, Cut, dt, eps, lam, Name)
                
                #Add Temp, PE, and time to lists at desired frequency, also print energies and temp to file
                if step%PlotFreq==0:    
                    TInst=InstTemp(Vel)
                    PEList.append(PEnergy)
                    TempList.append(TInst)
                    #print to file
                    #f.write("%8.6f \t %8.3f \t %8.3f \t %8.3f \t %8.3f \n" %((dt*step), TInst, PEnergy, KEnergy,PEnergy+KEnergy))
                    
                #Update realtime visual if being used
                if UseVisual:
                    if atomvis.Initialized:
                        #update the positions
                        atomvis.Update(Pos)
                    else:
                        #initialize the visualization window
                        atomvis.Init(Pos)
                        
            """Second Equilibration Run"""
            for step in range(NStepsEquil2):
                #Check if need thermostat
                if step%AndersenFreq==0:
                    Vel=Andersen(Vel,T)
                
                #Do MD Step
                Pos, Vel, Accel, KEnergy, PEnergy, CollForces = prolib.vvintegrate(Pos, Vel, Accel, M, L, Cut, dt, eps, lam, Name)
                
                #Add total energy to list
                EList.append(PEnergy+KEnergy)
                
                #Add to lists at desired frequency
                if step%PlotFreq==0:
                    Tinst=InstTemp(Vel)
                    PEList.append(PEnergy)
                    TempList.append(Tinst)
                    #Write energies and temp to file if want
                    #f.write("%8.6f \t %8.3f \t %8.3f \t %8.3f \t %8.3f \n" %((dt*step), Tinst, PEnergy, KEnergy,PEnergy+KEnergy))
                
                #Update real time visual
                if UseVisual:
                    if atomvis.Initialized:
                        #update the positions
                        atomvis.Update(Pos)
                    else:
                        #initialize the visualization window
                        atomvis.Init(Pos)
            
            #Calculate average total energy            
            AverageEnergy=np.mean(EList)
            
            #Rescale velocities so total energy is average total energy from equilibration 2
            k2=2*(AverageEnergy-PEnergy)/np.sum(Vel*Vel)
            #find rescaling factor    
            k=np.sqrt(k2)
            #rescale velocities
            Vel=k*Vel
            
            #Lists for production run
            EndtoEndList=[]     #End to End Distance List for this run
            Forces1List=[]      #Forces on colloid 1 list for this run
            Forces2List=[]      #Forces on colloid 2 list for this run
            
            """Production Run"""
            for step in range(NStepsProd):
                #If want Andersen on for production, uncomment next two lines
                #if step%AndersenFreq==0:
                    #Vel=Andersen(Vel,T)
                
                #do the verlet integrator to get the new Pos, Vel, Accel, KEnergy, and PEnergy        
                Pos, Vel, Accel, KEnergy, PEnergy, CollForces = prolib.vvintegrate(Pos, Vel, Accel, M, L, Cut, dt, eps, lam, Name)
                
                #Add end to end distance to list
                EndtoEndList.append(EndtoEnd(Pos))
                
                #After 1 second, start tracking forces
                if step>10000:
                    Forces1List.append(CollForces[0,0])
                    Forces2List.append(CollForces[1,0])
                
                #Update other lists at desired frequency
                if step%PlotFreq==0:
                    Tinst=InstTemp(Vel)
                    PEList.append(PEnergy)
                    TempList.append(Tinst)
                    #f.write("%8.6f \t %8.3f \t %8.3f \t %8.3f \t %8.3f \n" %((dt*step), Tinst, PEnergy, KEnergy,PEnergy+KEnergy))
                    #Write the time, Tinst, PEnergy, and KEnergy to the file if desired
                
                #update positions on visual if using        
                if UseVisual:
                    if atomvis.Initialized:
                        #update the positions
                        atomvis.Update(Pos)
                    else:
                        #initialize the visualization window
                        atomvis.Init(Pos)
                
                #If want to make movie, leave next three steps uncommented
                #if MovieStartStep<step<MovieStopStep:
                    #if step%PdbFreq==0:
                        #Pdb.write(Pos)
            
            f.write("%i \t %i \t %i \t %8.3f \t %8.3f \n" %(M, CollDist, NTrial, np.mean(Forces1List), np.sqrt(np.mean(EndtoEndList))))            
            #Add the mean end to end distance from this run to the list
            DistList.append(np.sqrt(np.mean(EndtoEndList)))
            
            #Plot the potential energy versus time for this run if want to check equilibration
            #plt.plot(timelist,PEList,colorlist[colorlistindex],label="%i" %CollDist)
            
            #Add the mean force on colloid 1 to the mean force list
            MeanForceList.append(np.mean(Forces1List))
            
            #Advance the color index for the PE plot
            #colorlistindex=colorlistindex+1
            
            #Add the list of temperatures from this run to the main list of temperatures
            #MainTempList.append(TempList)

#Add legend to PE plot
#plt.legend(loc=2)
#Add labels
#plt.xlabel("Time")
#plt.ylabel("Potential Energy")

#Plot end to end distance versus colloid separation distance
#plt.figure()
#plt.scatter(CollDistList,DistList)
#plt.xlabel("Colloid Separation Distance")
#plt.ylabel("Polymer End-to-End Distance")

#Plot mean force on colloid 1 versus separation distance
#plt.figure()
#plt.scatter(CollDistList,MeanForceList)
#plt.xlabel("Colloid Separation Distance")
#plt.ylabel("Mean Force")

#Plot temperature versus time
#MainTempList=np.array(MainTempList)

#for i in range(len(MainTempList)):
    #plt.figure()
    #plt.scatter(timelist,MainTempList[i],label="%i" %CollDistList[i])
    #plt.legend(loc=2)


#close files
f.close()
#Pdb.close()




